@@ -1,14 +1,14 @@
 /* Generate from machine description:
    - some flags HAVE_... saying which simple standard instructions are
    available for this machine.
-   Copyright (C) 1987, 1991, 1995, 1998,
-   1999, 2000, 2003, 2004 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1991, 1995, 1998, 1999, 2000, 2003, 2004, 2007
+   Free Software Foundation, Inc.
 
 This file is part of GCC.
 
 GCC is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
-Software Foundation; either version 2, or (at your option) any later
+Software Foundation; either version 3, or (at your option) any later
 version.
 
 GCC is distributed in the hope that it will be useful, but WITHOUT ANY
@@ -17,9 +17,8 @@ FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
-along with GCC; see the file COPYING.  If not, write to the Free
-Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
-02110-1301, USA.  */
+along with GCC; see the file COPYING3.  If not see
+<http://www.gnu.org/licenses/>.  */
 
 
 #include "bconfig.h"
@@ -44,7 +43,7 @@ static void max_operand_1 (rtx);
 static int num_operands (rtx);
 static void gen_proto (rtx);
 static void gen_macro (const char *, int, int);
-static void gen_insn (rtx);
+static void gen_insn (int, rtx);
 
 /* Count the number of match_operand's found.  */
 
@@ -188,13 +187,32 @@ gen_proto (rtx insn)
 }
 
 static void
-gen_insn (rtx insn)
+gen_insn (int line_no, rtx insn)
 {
   const char *name = XSTR (insn, 0);
   const char *p;
+  const char *lt, *gt;
   int len;
   int truth = maybe_eval_c_test (XSTR (insn, 2));
 
+  lt = strchr (name, '<');
+  if (lt && strchr (lt + 1, '>'))
+    {
+      message_with_line (line_no, "unresolved iterator");
+      have_error = 1;
+      return;
+    }
+
+  gt = strchr (name, '>');
+  if (lt || gt)
+    {
+      message_with_line (line_no,
+			 "unmatched angle brackets, likely "
+			 "an error in iterator syntax");
+      have_error = 1;
+      return;
+    }
+
   /* Don't mention instructions whose names are the null string
      or begin with '*'.  They are in the machine description just
      to be recognized.  */
@@ -261,7 +279,7 @@ main (int argc, char **argv)
       if (desc == NULL)
 	break;
       if (GET_CODE (desc) == DEFINE_INSN || GET_CODE (desc) == DEFINE_EXPAND)
-	gen_insn (desc);
+	gen_insn (line_no, desc);
     }
 
   /* Print out the prototypes now.  */
@@ -274,7 +292,7 @@ main (int argc, char **argv)
 
   puts("\n#endif /* GCC_INSN_FLAGS_H */");
 
-  if (ferror (stdout) || fflush (stdout) || fclose (stdout))
+  if (have_error || ferror (stdout) || fflush (stdout) || fclose (stdout))
     return FATAL_EXIT_CODE;
 
   return SUCCESS_EXIT_CODE;
