@@ -1802,7 +1802,7 @@ NAME (TYPE x, int m)
 #define isfinite(x)	__builtin_expect (!isnan((x) - (x)), 1)
 #define isinf(x)	__builtin_expect (!isnan(x) & !isfinite(x), 0)
 
-#define INFINITY	CONCAT2(__builtin_inf, CEXT) ()
+#define INFINITY	CONCAT2(__builtin_huge_val, CEXT) ()
 #define I		1i
 
 /* Helpers to make the following code slightly less gross.  */
@@ -2049,6 +2049,8 @@ __enable_execute_stack (void *addr __attribute__((__unused__)))
 /* Jump to a trampoline, loading the static chain address.  */
 
 #if defined(WINNT) && ! defined(__CYGWIN__)
+int getpagesize (void);
+int mprotect (char *,int, int);
 
 int
 getpagesize (void)
